#include "tap.h"
#include "string.h"

#define TAPNAME "DebugCmds v1.1"

TAP_ID(0x8beef006);
TAP_PROGRAM_NAME(TAPNAME);
TAP_AUTHOR_NAME("DeadBeef");
TAP_DESCRIPTION("Special commands");
TAP_ETCINFO(__DATE__);

#define RX_BUF_SIZE 4096

// interrupt vector related definitions
#define RX_INT_VEC 0x31
// hardware register for serial IO
#define SIO_STAT_REG 0xb2002050
#define SIO_OUT_REG 0xb2002000

typedef void (*funcVoid)(void);
typedef int (*funcCharPtr)(char*);

static char rxBuffer[RX_BUF_SIZE];
static int wIndex = 0;
static int rIndex = 0;
static funcVoid rxIsrBackup = NULL;
static dword *pVecTableBase = NULL;

#define JAL_CMD 0x0c000000
#define JR_CMD 0x03e00008
#define JAL_MASK 0xff000000

typedef void (*CmdParser)(char* cmdLine);
CmdParser cmdParser = NULL;

//-----------------------------------------------------------------------------
// initCodeWrapper() creates a code wrapper based on TAP_GetTime. The function
// allocates a buffer and copies the code of TAP_GetTime into it except for
// the actual call of the getTime() firmware function. The address of the 
// getTime() is replaced with the provided address.
// The wrapped function call supports up to 4 parameters and returns a
// short value.
// Parameters:
//      fwFuncAddr - address of the FW function to be wrapped
// initCodeWrapper() returns a buffer with code wrapping the call to provided
// firmware function address.
// The buffer can be assigned to a function pointer which can be used to execute
// the wrapped code.
dword* initCodeWrapper(dword fwFuncAddr)
{
  int i;
  // TAP_GetTime provides a template for wrapper code. The only thing to be
  // changed is the address of the wrapped function.
  dword *pSrc = (dword*)TAP_GetTime;
  dword *pBuffer;
  int bufSize = 0;

  // find the return statement
  while(pSrc[bufSize] != JR_CMD)
  {
    bufSize++;
  }

  // the buffer should include the return statement and the command following
  // the return statement
  bufSize += 2;

  // allocate memory
  pBuffer = malloc(sizeof(dword) * bufSize);

  if(pBuffer == NULL)
  {
    return NULL;
  }

  for(i = 0; i < bufSize; i++, pSrc++)
  {
    if((*pSrc & JAL_MASK) == JAL_CMD)
    {
      // this is the wrapped call to the actual FW implementation
      // replace it with the call to the provided address
      pBuffer[i] = JAL_CMD | ((fwFuncAddr & ~JAL_MASK) >> 2);
    }
    else
    {
      pBuffer[i] = *pSrc;
    }
  }

  return pBuffer;
}

//-----------------------------------------------------------------------------
// This function sets the interrupt service routine
// for the specified vector and returns the address
// of the old interrupt service routine.
void* setIntVector(int vector, void *pFunc)
{
  void *pRet;

  if((vector < 0) || (vector > 63))
    return NULL;

  pRet = (void*)pVecTableBase[vector];
  pVecTableBase[vector] = (dword)pFunc;

  return pRet;
}

//-----------------------------------------------------------------------------
// This function handles interrupts from serial Rx.
void rxIntHandler()
{
  // The pointers have to be volatile in order to avoid
  // usage of $gp register in interrupt context.
  // The problem is that $gp value in interrupt context
  // differs from the value in TAP context.
  // The compiler produces code using $ra for TAP base
  // address. Though, it is possible that the $ra contains
  // wrong address if an interrupt occurs while another TAP
  // is executed.
  // TODO: implement a global pointer with constant address
  //       to access TAP data
  volatile dword *pStatus = (dword*)SIO_STAT_REG;
  volatile dword *pData = (dword*)SIO_OUT_REG;
  volatile int *pRIndex = &rIndex;
  volatile int *pWIndex = &wIndex;
  volatile char *pRxBuffer = rxBuffer;

  // read from receiver FIFO until empty
  while(*pStatus & 0x00000001)
  {
    // read character to buffer
    pRxBuffer[*pWIndex] = *pData;

    // store character to buffer
    *pWIndex = (*pWIndex + 1) % RX_BUF_SIZE;

    // check for overflow
    if(*pWIndex == *pRIndex)
    {
      // increment the rIndex
      *pRIndex = (*pRIndex + 1) % RX_BUF_SIZE;
    }
  }
}

//-----------------------------------------------------------------------------
// This function restores the original interrupt service routine
// and terminates the TAP.
void exitTap()
{
  // restore interrupt vector
  setIntVector(RX_INT_VEC, rxIsrBackup);

  // free the code wrapper buffer
  free(cmdParser);

  // print exit message
  TAP_Print("\nDebug TAP terminated\n");
  TAP_Exit();
}

//-----------------------------------------------------------------------------
// This function processes the input buffer to get a line
// terminated by a carrige return (CR).
// On occurence of CR the line is passed to the extended
// command handler in the firmware.
void processInput()
{
  static char line[255];
  static int length = 0;
  char ch;

  while(rIndex != wIndex)
  {
    ch = rxBuffer[rIndex];

    // check the line length
    if(length >= (sizeof(line) - 1))
    {
      TAP_Print("\nCommand line too long!\n");
      length = 0;

      // display new prompt
      TAP_Print("\n=> ");
    }

    switch(ch)
    {
    case '\r':
      // CR found
      if(!length)
      {
        // display new prompt
        TAP_Print("\n=> ");

        break;
      }

      // terminate string
      line[length] = 0;

      TAP_Print("\n");

      // check for exit command
      if(strcmp(line, "exit") == 0)
      {
        exitTap();
        return;
      }

      // otherwise call command handler
      cmdParser(line);

      // display new prompt
      TAP_Print("\n=> ");

      // reset line
      length = 0;
      break;
    case '\b':
      // backspace
      if(length > 0)
      {
        TAP_Print("\b \b");
        length--;
      }
      break;
    default:
      if((ch < 0x20) || (ch > 0x7e))
      {
        // ignore character
        break;
      }

      // display character
      TAP_PutCh(ch);

      // convert lower case to upper case
      if((ch > 0x60) && (ch < 0x7b))
      {
        //ch -= 0x20;
      }

      // copy character to line buffer
      line[length] = ch;
      length++;
    }

    rIndex = (rIndex + 1) % RX_BUF_SIZE;
  }
}

//-----------------------------------------------------------------------------
dword TAP_EventHandler( word event, dword param1, dword param2 )
{
  if(event == EVT_IDLE)
  {
    // process input from the Rx buffer, if any
    processInput();
  }

  if((event == EVT_KEY) && (param1 == RKEY_Exit))
  {
    //exitTap();
  }

  return param1;
}


//-----------------------------------------------------------------------------
int TAP_Main(void)
{
  dword cmdParserAddr ;

  extern int _appl_version;
  switch(_appl_version)
  {
  case  0x1155:
    // adresses for the FW 03/24/05 (TF5000PVR)
    // vector table base
    pVecTableBase = (dword*)0x81b12154;

    // address of the built-in command parser for additional commands
    cmdParserAddr = 0x8010b7b0;
    break;
  case  0x1188:
    // adresses for the FW 06/10/05 (TF5000PVR)
    // vector table base
    pVecTableBase = (dword*)0x81b12154;

    // address of the built-in command parser for additional commands
    cmdParserAddr = 0x8010b470;
    break;
  default:
    TAP_Print("\nDebugCmds TAP: Unsupported firmware version\n");
    TAP_Print("DebugCmds TAP only supports firmware versions 5.11.55 and 5.11.88\n");
    return 0;
  }

  cmdParser = (CmdParser)initCodeWrapper(cmdParserAddr);
  if(cmdParser == NULL)
  {
    TAP_Print("wrapper initialization failed\n");
    return 0;
  }

  // set interrupt vector
  rxIsrBackup = setIntVector(RX_INT_VEC,rxIntHandler);

  // print start message
  TAP_Print("\nDebug TAP started\n\n");
  TAP_Print("=> ");

  return 1;
}
